import params
from api import API


def describe_instances(instance_ids=None,
                       status=None,
                       search_word=None,
                       limit=None,
                       offset=None,
                       reverse=None,
                       verbose=None):

        payload = {}
        payload = params.filter_ids(payload, 'instanceIds', instance_ids)
        payload = params.filter_status(payload, status)
        payload = params.filter_search_word(payload, search_word)
        payload = params.filter_limit(payload, limit)
        payload = params.filter_offset(payload, offset)
        payload = params.filter_reverse(payload, reverse)
        payload = params.filter_verbose(payload, verbose)

        instances = API.conn.call('DescribeInstances', payload)
        return instances



from requests.structures import LookupDict

_codes = {
    0: ('ok',),
    4100: ('param_invalid',),
    4101: ('access_key_invalid',),
    4103: ('resource_unauthorized',),
    4104: ('resource_not_found',),
    4105: ('action_disallow',),
    4110: ('request_format_invalid',),
    4113: ('quota_not_enough',),
    5000: ('server_error',),
    5001: ('actions_totally_failed',),
    5002: ('actions_partially_failed',),
}

codes = LookupDict(name='ret_codes')

for (code, titles) in list(_codes.items()):
    for title in titles:
        setattr(codes, title, code)



import Queue
from gevent.coros import BoundedSemaphore
from instance import find_proper_flavor

from plato_cat.cases import wait_for_job
from plato_cat.cases import get_resource
from plato_cat.cases import wait_busy_resource
from plato_cat.cases import post_resource


class VolumeCase():

    def run(self, API, sleep):
        result = API.conn.call('CreateVolumes', {
            'size': 1,
            'name': 'create-volume-name',
            'volumeType': 'normal',
            'count': 2
        })

        job_id = result['jobId']
        wait_for_job(API, job_id, sleep)

        volume_ids = result['volumeIds']
        volumes = get_resource(API, 'DescribeVolumes', {
            'status': ['active'],
            'volumeIds': volume_ids
        }, 'volumeSet')

        for v in volumes:
            if v['status'] != 'active':
                raise Exception('volume (%s) creation job finished,'
                                ' but status is not [active], but [%s]'
                                % (v['volumeId'], v['status']))

        instances = get_resource(API, 'DescribeInstances', {
            'status': ['active', 'stopped', 'error']
        }, 'instanceSet')
        instance_ids = [i['instanceId'] for i in instances]

        # attach volume
        result = API.conn.call('AttachVolume', {
            'instanceId': instance_ids[0],
            'volumeId': volume_ids[0],
        })

        job_id = result['jobId']
        wait_for_job(API, job_id, sleep)

        # extend volumes
        result = API.conn.call('ExtendVolumes', {
            'volumeIds': [volume_ids[1]],
            'size': 2,
        })

        job_id = result['jobId']
        wait_for_job(API, job_id, sleep)


class CleanVolumes():

    def run(self, API, sleep):
        wait_busy_resource(API, 'DescribeVolumes', {
            'status': ['pending', 'attaching', 'detaching',
                       'backup_ing', 'backup_restoring'],
        }, sleep)

        volumes = get_resource(API, 'DescribeVolumes', {
            'status': ['active', 'inuse', 'error']
        }, 'volumeSet')

        # detach volumes
        job_ids = []
        for volume in volumes:
            if volume['instanceId']:
                result = API.conn.call('DetachVolumes', {
                    'volumeIds': [volume['volumeId']],
                    'instanceId': volume['instanceId']
                })

                job_ids.append(result['jobId'])

        wait_for_job(API, job_ids, sleep)

        volume_ids = [v['volumeId'] for v in volumes]
        post_resource(API, 'DeleteVolumes', volume_ids, 'volumeIds')


class BenchVolumes():

    def __init__(self, API, concurrent, sleep):
        result = API.conn.call('CreateNetwork', {
            'name': 'network-for-bench-volumes',
        })

        job_id = result['jobId']
        network_id = result['networkId']

        # wait network creation finish
        wait_for_job(API, job_id, sleep)

        subnet_id = API.conn.call('CreateSubnet', {
            'networkId': network_id,
            'cidr': '192.168.200.0/24',
        })['subnetId']

        instance_types = get_resource(API, 'DescribeInstanceTypes', {
            'status': ['active']
        }, 'instanceTypeSet')
        selected_instance_type = find_proper_flavor(
            instance_types, 2048, 1, 40)

        images = get_resource(API, 'DescribeImages', {
            'status': ['active'],
            'isPublic': True,
        }, 'imageSet')
        selected_image = images[0]

        # create the instance and wait for it to be done.
        result = API.conn.call('CreateInstances', {
            'name': 'instance-for-bench-volumes',
            'imageId': selected_image['imageId'],
            'instanceTypeId': selected_instance_type['instanceTypeId'],
            'subnetId': subnet_id,
            'loginMode': 'password',
            'loginPassword': 'chenjie1234',
            'count': concurrent
        })
        job_id = result['jobId']
        wait_for_job(API, job_id, sleep)

        self.instance_ids_queue = Queue.Queue()
        [self.instance_ids_queue.put(instance_id)
            for instance_id in result['instanceIds']]

        self.lock = BoundedSemaphore(1)
        self.counter = 0

    def run(self, API, sleep):
        # do a complete process:
        # step 1. create volume
        # step 2. attach volume
        # step 3. detach volume
        # step 4. delete volume
        try:
            self.lock.acquire()
            self.counter += 1
            self.lock.release()

            instance_id = self.instance_ids_queue.get()

            result = API.conn.call('CreateVolumes', {
                'name': 'bench-name-%s' % self.counter,
                'size': 1, 'count': 1, 'volumeType': 'normal'
            })

            job_id = result['jobId']
            wait_for_job(API, job_id, sleep)

            volume_id = result['volumeIds'][0]

            # attach volume
            result = API.conn.call('AttachVolume', {
                'instanceId': instance_id,
                'volumeId': volume_id,
            })

            job_id = result['jobId']
            wait_for_job(API, job_id, sleep)

            result = API.conn.call('DetachVolumes', {
                'volumeIds': [volume_id],
                'instanceId': instance_id
            })

            job_id = result['jobId']
            wait_for_job(API, job_id, sleep)

            post_resource(API, 'DeleteVolumes', [volume_id], 'volumeIds')

            return True

        except Exception as ex:
            print ex
            return False

        finally:
            self.instance_ids_queue.put(instance_id)




class ConsoleAlert(object):

    def call(self, exceptions):
        print " \nexceptions RAISED!! : \n"
        for ex in exceptions:
            print 'case: ', ex['case']
            print 'message: ', ex['message']
            print 'traceback: ', ex['traceback']

    def report(self, title, is_pass, stat):
        print title
        print 'Pass: ', is_pass
        print 'Stat: ', stat



import json
import requests
from codes import codes
import errors


class Conn(object):
    def __init__(self, access_key, access_secret, endpoint, is_debug=False):
        self.access_key = access_key
        self.access_secret = access_secret
        self.endpoint = endpoint

    def make_headers(self):
        headers = {
            'Content-Type': 'application/json; charset=utf-8',
            'X-Le-Key': self.access_key,
            'X-Le-Secret': self.access_secret,
        }
        return headers

    def call(self, action, payload={}):
        """
        make a action call.
        """
        url = self.endpoint
        headers = self.make_headers()

        payload['action'] = action
        res = requests.post(url, headers=headers, json=payload)

        data = self.parse_response(action, res)
        return data

    def parse_response(self, action, res):
        if res.status_code != 200:
            raise errors.HTTPError(res.status_code)

        ret = json.loads(res.text)
        ret_code = ret['retCode']
        data = ret['data']

        if ret_code == codes.ok:
            print action, 'ok'
            return data

        message = ret['message'] + '\n exceptionTag: ' + data['exceptionTag']

        if ret_code == codes.param_invalid:
            raise errors.ParamInvalidError(message)

        elif ret_code == codes.access_key_invalid:
            raise errors.AccessKeyInvalidError(message)

        elif ret_code == codes.resource_unauthorized:
            raise errors.ResourceUnauthorizedError(message)

        elif ret_code == codes.resource_not_found:
            raise errors.ResourceNotFoundError(message)

        elif ret_code == codes.action_disallow:
            raise errors.ActionDisallowError(message)

        elif ret_code == codes.request_format_invalid:
            raise errors.RequestFormatInvalidError(message)

        elif ret_code == codes.quota_not_enough:
            raise errors.QuotaNotEnoughError(message)

        elif ret_code == codes.server_error:
            raise errors.ServerErrorError(message)

        elif ret_code == codes.actions_totally_failed:
            raise errors.ActionsTotallyFailedError(message)

        elif ret_code == codes.actions_partially_failed:
            raise errors.ActionsPartiallyFailedError(message)

        else:
            raise Exception(message)



import params
from api import API


def describe_volumes(volume_ids=None,
                     status=None,
                     search_word=None,
                     limit=None,
                     offset=None,
                     reverse=None,
                     verbose=None):

        payload = {}
        payload = params.filter_ids(payload, 'volumeIds', volume_ids)
        payload = params.filter_status(payload, status)
        payload = params.filter_search_word(payload, search_word)
        payload = params.filter_limit(payload, limit)
        payload = params.filter_offset(payload, offset)
        payload = params.filter_reverse(payload, reverse)
        payload = params.filter_verbose(payload, verbose)

        volumes = API.conn.call('DescribeVolumes', payload)
        return volumes





def filter_ids(payload, ids_key, ids):
    if ids:
        payload[ids_key] = ids
    return payload


def filter_status(payload, status):
    if status:
        payload['status'] = status
    return payload


def filter_search_word(payload, search_word):
    if search_word:
        payload['searchWord'] = search_word
    return payload


def filter_limit(payload, limit):
    if limit:
        payload['limit'] = limit
    return payload


def filter_offset(payload, offset):
    if offset:
        payload['offset'] = offset
    return payload


def filter_reverse(payload, reverse):
    if reverse is not None:
        payload['reverse'] = reverse
    return payload


def filter_verbose(payload, verbose):
    if verbose is not None:
        payload['verbose'] = verbose
    return payload



from plato_cat.cases import wait_for_job
from plato_cat.cases import get_resource
from plato_cat.cases import wait_busy_resource
from plato_cat.cases import post_resource


class NetworkCase():

    def run(self, API, sleep):
        result = API.conn.call('CreateNetwork', {
            'name': 'network-creation-case',
        })

        network_id = result['networkId']
        job_id = result['jobId']

        wait_for_job(API, job_id, sleep)

        networks = get_resource(API, 'DescribeNetworks', {
            'status': ['active'],
            'networkIds': [network_id]
        }, 'networkSet')

        for n in networks:
            if n['status'] != 'active':
                raise Exception('network (%s) creation job finished,'
                                ' but status is not [active], but [%s]'
                                % (n['networkId'], n['status']))

        API.conn.call('CreateSubnet', {
            'networkId': network_id,
            'cidr': '192.168.200.0/24',
        })['subnetId']

        subnets = get_resource(API, 'DescribeSubnets', {
            'networkIds': [network_id],
            'status': ['active']
        }, 'subnetSet')

        if len(subnets) != 1:
            raise Exception('create subnet failed.')

        # set external gateway
        result = API.conn.call('SetExternalGateway', {
            'networkIds': [network_id],
        })
        result['networkIds']

        # update external gateway bandwidth
        API.conn.call('UpdateExternalGatewayBandwidth', {
            'networkIds': [network_id],
            'bandwidth': 1,
        })

        # create port forwarding
        API.conn.call('CreatePortForwarding', {
            'networkId': network_id,
            'protocol': 'tcp',
            'outsidePort': 10086,
            'insideAddress': '192.168.200.3',
            'insidePort': 10010,
        })


class CleanNetworks():

    def run(self, API, sleep):
        port_forwardings = get_resource(API, 'DescribePortForwardings', {
            'status': ['active']
        }, 'portForwardingSet')

        port_forwarding_ids = [i['portForwardingId']
                               for i in port_forwardings]

        post_resource(API, 'DeletePortForwardings',
                      port_forwarding_ids, 'portForwardingIds')

        wait_busy_resource(API, 'DescribeNetworks', {
            'status': ['pending', 'building'],
        }, sleep)

        networks = get_resource(API, 'DescribeNetworks', {
            'status': ['active', 'disabled', 'error']
        }, 'networkSet')

        network_ids = [n['networkId']
                       for n in networks if n['externalGatewayIp']]

        # unset external gateway
        post_resource(API, 'UnsetExternalGateway',
                      network_ids, 'networkIds')

        subnets = get_resource(API, 'DescribeSubnets', {
            'status': ['active']
        }, 'subnetSet')

        subnet_ids = [i['subnetId'] for i in subnets]

        post_resource(API, 'DeleteSubnets', subnet_ids, 'subnetIds')

        # delete networks
        network_ids = [n['networkId'] for n in networks]
        post_resource(API, 'DeleteNetworks', network_ids, 'networkIds')



from plato_cat.cases import get_resource
from plato_cat.cases import post_resource


class EIPCase():

    def run(self, API, sleep):
        eips = get_resource(API, 'DescribeEips', {
            'status': ['active']
        }, 'eipSet')

        eip_ids = API.conn.call('AllocateEips', {
            'bandwidth': 1,
            'count': 1
        })['eipIds']

        new_eips = get_resource(API, 'DescribeEips', {
            'status': ['active']
        }, 'eipSet')

        if len(new_eips) - len(eips) != 1:
            raise Exception("after Allocate 1 Eips,"
                            "DescribeEips did not increase by 1!")

        API.conn.call('UpdateBandwidth', {
            'eipIds': eip_ids,
            'bandwidth': 1
        })

        instances = get_resource(API, 'DescribeInstances', {
            'status': ['active', 'stopped', 'error']
        }, 'instanceSet')
        instance_ids = [i['instanceId'] for i in instances]

        API.conn.call('AssociateEip', {
            'eipId': eip_ids[0],
            'instanceId': instance_ids[0],
        })


class CleanEips():

    def run(self, API, sleep):
        eips = get_resource(API, 'DescribeEips', {
            'status': ['active', 'associated'],
        }, 'eipSet')

        # dissociate eips
        eip_ids = [e['eipId'] for e in eips if e['resourceId']]
        post_resource(API, 'DissociateEips', eip_ids, 'eipIds')

        eip_ids = [e['eipId'] for e in eips]
        post_resource(API, 'ReleaseEips', eip_ids, 'eipIds')



from plato_cat.cases import get_resource
from plato_cat.cases import wait_busy_resource
from plato_cat.cases import post_resource


class ImageCase():

    def run(self, API, sleep):
        API.conn.call('DescribeImages')


class CleanImages():

    def run(self, API, sleep):
        wait_busy_resource(API, 'DescribeImages', {
            'status': ['pending'],
            'isPublic': False
        }, sleep)

        images = get_resource(API, 'DescribeImages', {
            'status': ['active', 'error'],
            'isPublic': False
        }, 'imageSet')

        image_ids = [i['imageId'] for i in images]
        post_resource(API, 'DeleteImages', image_ids, 'imageIds')



import os
import requests
import json

from oslo_config import cfg
CONF = cfg.CONF

Incoming = os.getenv('PLATO_CAT_SLACK_HOOK')


class SlackAlert(object):

    def call(self, exceptions):
        if not Incoming:
            return

        endpoint = CONF.endpoint
        access_key = CONF.key

        for exception in exceptions:
            url = Incoming
            payload = json.dumps({
                'text': 'endpoint: %s\naccess key: %s\n'
                'case: %s\nexception: %s\ntraceback: %s'
                % (endpoint, access_key, exception['case'],
                   exception['message'], exception['traceback']),
            })

            try:
                requests.post(url, payload)
            except:
                pass

    def report(self, title, is_pass, stat):
        if not Incoming:
            return

        url = Incoming

        text = [title, 'Pass: %s' % is_pass]
        text += ['%s: %s' % (k, v) for k, v in stat.items()]
        text = '\n'.join(text)
        payload = json.dumps({'text': text})

        try:
            requests.post(url, payload)
        except:
            pass



from plato_cat.cases import wait_for_job
from plato_cat.cases import get_resource
from plato_cat.cases import wait_busy_resource
from plato_cat.cases import post_resource


def find_proper_flavor(instance_types, memory, vcpus, disk):
    for instance_type in instance_types:
        if instance_type['memory'] == memory and \
           instance_type['vcpus'] == vcpus and \
           instance_type['disk'] == disk:
            return instance_type


class InstanceCase():

    def run(self, API, sleep):
        instances = get_resource(API, 'DescribeInstances', {
            'status': ['active']
        }, 'instanceSet')

        result = API.conn.call('CreateNetwork', {
            'name': 'network-for-instance1',
        })

        job_id = result['jobId']
        network_id = result['networkId']

        # wait network creation finish
        wait_for_job(API, job_id, sleep)

        subnet_id = API.conn.call('CreateSubnet', {
            'networkId': network_id,
            'cidr': '192.168.200.0/24',
        })['subnetId']

        # set external gateway
        result = API.conn.call('SetExternalGateway', {
            'networkIds': [network_id],
        })

        instance_types = get_resource(API, 'DescribeInstanceTypes', {
            'status': ['active']
        }, 'instanceTypeSet')

        images = get_resource(API, 'DescribeImages', {
            'status': ['active'],
            'isPublic': True,
        }, 'imageSet')
        selected_image = images[0]

        # select proper instance_type & image
        selected_instance_type = find_proper_flavor(
            instance_types, 2048, 1, 40)

        if not selected_image:
            raise Exception('no proper image or proper instance_type '
                            'found to create instance')

        # create the instance and wait for it to be done.
        result = API.conn.call('CreateInstances', {
            'name': 'created-instance',
            'imageId': selected_image['imageId'],
            'instanceTypeId': selected_instance_type['instanceTypeId'],
            'subnetId': subnet_id,
            'loginMode': 'password',
            'loginPassword': 'chenjie1234',
            'count': 3
        })
        job_id = result['jobId']
        instance_ids = result['instanceIds']

        wait_for_job(API, job_id, sleep)
        # double check.
        instances = get_resource(API, 'DescribeInstances', {
            'instanceIds': instance_ids
        }, 'instanceSet')

        for i in instances:
            if i['status'] != 'active':
                raise Exception('instance (%s) creation job finished,'
                                ' but status is not [active], but [%s]'
                                % (i['instanceId'], i['status']))

        instance_id_1 = instance_ids[0]
        instance_id_2 = instance_ids[1]
        instance_id_3 = instance_ids[2]

        # stop all instances:  NO 1. NO 2. NO 3.
        result = API.conn.call('StopInstances', {
            'instanceIds': instance_ids,
        })
        job_id = result['jobId']

        wait_for_job(API, [job_id], sleep)

        # start instances  NO 1.
        result = API.conn.call('StartInstances', {
            'instanceIds': [instance_id_1],
        })
        job_id_1 = result['jobId']

        # reset instances  NO 2.
        result = API.conn.call('ResetInstances', {
            'instanceIds': [instance_id_2],
            'loginMode': 'password',
            'loginPassword': 'chenjie1234',
        })
        job_id_2 = result['jobId']

        # resize instances NO 3.
        instance_type_new = find_proper_flavor(
            instance_types, 2048, 2, 40)
        result = API.conn.call('ResizeInstances', {
            'instanceIds': [instance_id_3],
            'instanceTypeId': instance_type_new['instanceTypeId'],
        })
        job_id_3 = result['jobId']

        wait_for_job(API, [job_id_1, job_id_2, job_id_3], sleep)

        # after reset, instance may enter active status. stop it.
        double_check = get_resource(API, 'DescribeInstances', {
             'instanceIds': [instance_id_2]
        }, 'instanceSet')[0]['status']
        if double_check == 'active':
            result = API.conn.call('StopInstances', {
                'instanceIds': [instance_id_2],
            })
            job_id = result['jobId']
            wait_for_job(API, job_id, sleep)

        #
        # from here, instance NO 1 active, NO 2 stopped, NO 3 stopped.
        #

        # restart instances NO 1.
        result = API.conn.call('RestartInstances', {
            'instanceIds': [instance_id_1],
        })
        job_id_1 = result['jobId']
        # start instance NO 2 and NO 3.
        result = API.conn.call('StartInstances', {
            'instanceIds': [instance_id_2, instance_id_3],
        })
        job_id_2 = result['jobId']

        wait_for_job(API, [job_id_1, job_id_2], sleep)

        # capture instance
        result = API.conn.call('CaptureInstance', {
            'instanceId': instance_id_1,
            'name': 'test-capture-instance',
        })
        job_id = result['jobId']

        wait_for_job(API, job_id, sleep)

        # connect VNC
        result = API.conn.call('ConnectVNC', {
            'instanceId': instance_id_2,
        })
        # get instance output
        result = API.conn.call('GetInstanceOutput', {
            'instanceId': instance_id_2,
        })


class CleanInstances():

    def run(self, API, sleep):
        wait_busy_resource(API, 'DescribeInstances', {
            'status': ['pending', 'starting', 'stopping',
                       'restarting', 'scheduling']
        }, sleep)

        instances = get_resource(API, 'DescribeInstances', {
            'status': ['active', 'stopped', 'error']
        }, 'instanceSet')

        # delete instances
        instance_ids = [i['instanceId'] for i in instances]
        post_resource(API, 'DeleteInstances', instance_ids, 'instanceIds')



import json
import requests

####################################
#
# TO BE DONE. TODO
#
####################################


class RegionManager(object):

    def __init__(self, regions_url):
        self.known_regions = {}
        self.url = regions_url
        self.init_regions()

    def init_regions(self):
        res = requests.get(self.url)
        data = json.loads(res.text)['data']
        for region in data['regions']:
            name = region['name']
            endpoint = region['endpoint']
            self.known_regions[name] = endpoint

    def get_endpoint(self, region_name):
        return self.known_regions[region_name]

regions = RegionManager('http://localhost/regions')



# from plato_cat.cases import wait_for_job
# from plato_cat.cases import get_resource
# from plato_cat.cases import post_resource


class SnapshotCase():

    def run(self, API, sleep):
        pass
#         snapshots = get_resource(API, 'DescribeSnapshots',
#                                       {}, 'snapshotSet')

#         volumes = get_resource(API, 'DescribeVolumes', {
#             'status': ['active']
#         }, 'volumeSet')

#         volume_ids = [v['volumeId'] for v in volumes]

#         result = API.conn.call('CreateSnapshots', {
#             'name': 'test-snapshot',
#             'volumeId': volume_ids[0],
#         })

#         job_id = result['jobId']

#         wait_for_job(API, job_id, sleep)


class CleanSnapshots():

    def run(self, API, sleep):
        pass
#         snapshots = get_resource(API, 'DescribeSnapshots',{
#             'status': ['active', 'error'],
#         }, 'snapshotSet')

#         snapshot_ids = [i['snapshotId'] for i in snapshots]

#         post_resource(API, 'DeleteSnapshots', snapshot_ids, 'snapshotIds')



import sys
import time


def get_resource(API, action, payload, result_set, max_length=400):
    """
    get max 400 length resources. that is 20 http requests round.
    """
    resources = []

    payload = payload or {}
    payload['limit'] = 20
    payload['offset'] = 0

    total = -1

    while True:
        payload['offset'] = len(resources)

        result = API.conn.call(action, payload)

        if total != result['total'] and total != -1:
            # total change. reset resources and start over.
            resources = []
            total = result['total']
            continue

        if total == -1:
            total = result['total']

        resources += result[result_set]

        if result['total'] <= len(resources):
            # have find all.
            break

        if len(resources) > max_length:
            # have reach max.
            break

    return resources


def wait_busy_resource(API, action, busy_payload, sleep, timeout=60*5):
    """
    describe busy resources, if total is not 0, means they are still busy,
    wait 4 seconds and try next time.
    """

    if timeout == 0:
        timeout = sys.maxint

    start = time.time()

    while True:
        result = API.conn.call(action, busy_payload)
        if result['total'] == 0:
            # no busy resource. good.
            break

        result.pop('total')
        result.pop('limit')
        result.pop('offset')

        id_key = result.keys()[0][0:-3] + 'Id'
        resources = result.values()[0]

        sleep(4)

        if time.time() - start > timeout:
            instance_status = dict((r[id_key], r['status']) for r in resources)
            raise Exception('wait_busy_resource timeout. '
                            'maybe job was not completely executed? '
                            'resources => status map: %s' % instance_status)


def post_resource(API, action, resources, ids_key):
    """
    post max 400 resources. that is 20 http requests round.
    """
    resources = resources[0:400]

    while resources:
        payload = {}
        payload[ids_key] = resources[:20]

        API.conn.call(action, payload)

        resources = resources[20:]


def wait_for_job(API, job_ids, sleep, timeout=60*5):
    """
    wait a job for at most 5 minutes.
    """
    if not job_ids:
        return
    if type(job_ids) is not list:
        job_ids = [job_ids]

    if timeout == 0:
        timeout = sys.maxint

    start = time.time()

    while True:
        job_set = API.conn.call('DescribeJobs', {
            'jobIds': job_ids
        })['jobSet']

        if len(job_set) == 0:
            sleep(2)
            continue

        busy = False

        busy_jobs = []
        for job in job_set:
            # if one is busy, go to next round
            if job['status'] in('pending', 'running'):
                busy = True
                busy_jobs.append(job)
                continue

            # if one job is failed, raise exception
            if job['status'] == 'error':
                raise Exception(
                    "Job (%s) failed, action: %s, resourceIds: %s" %
                    (job['jobId'], job['action'], job['resourceIds']))

        if not busy:
            break

        sleep(4)

        job_ids = [j['jobId'] for j in busy_jobs]

        if time.time() - start > timeout:
            job_status = dict((j['jobId'], j['status']) for j in busy_jobs)
            raise Exception('wait_for_job timeout. '
                            'job => status map: %s' % job_status)



from gevent import monkey
monkey.patch_all()  # noqa

import time
import sys
import traceback
import gevent
from gevent.pool import Pool
from gevent import Timeout
import functools

from sdk.actions import api as sdk_api

from cases.instance import InstanceCase
from cases.key_pair import KeyPairCase
from cases.eip import EIPCase
from cases.image import ImageCase
from cases.instance_type import InstanceTypeCase
from cases.job import JobCase
from cases.monitor import MonitorCase
from cases.network import NetworkCase
from cases.operation import OperationCase
from cases.quota import QuotaCase
from cases.snapshot import SnapshotCase
from cases.volume import VolumeCase

from cases.instance import CleanInstances
from cases.key_pair import CleanKeyPairs
from cases.image import CleanImages
from cases.network import CleanNetworks
from cases.volume import CleanVolumes
from cases.eip import CleanEips
from cases.snapshot import CleanSnapshots

from cases.volume import BenchVolumes

from alerts.slack import SlackAlert
from alerts.console import ConsoleAlert

from oslo_config import cfg
COMMON_OPTS = [
    cfg.StrOpt('endpoint', help='api endpoint'),
    cfg.StrOpt('key', help='access key'),
    cfg.StrOpt('secret', help='access secret')
]

BENCH_OPTS = [
    cfg.IntOpt('total', help=''),
    cfg.IntOpt('concurrent', help=''),
    cfg.IntOpt('ratio', help=''),
    cfg.IntOpt('timeout', help=''),
]

CONF = cfg.CONF
CONF.register_cli_opts(COMMON_OPTS)
CONF.register_opts(COMMON_OPTS)
CONF.register_opts(BENCH_OPTS, group='benchmark')


def _get_api(conf):
    if not (conf.key and conf.secret and conf.endpoint):
        raise Exception('You have to specifiy --key, --secret, --endpoint '
                        'in the cli params, or use config file with '
                        '--config-file /your/config/file/path, '
                        'check cat.conf.sample for example')

    return sdk_api.setup(access_key=conf.key,
                         access_secret=conf.secret,
                         endpoint=conf.endpoint,
                         is_debug=True)


def _run_case(case, api):
    try:
        case.run(api, gevent.sleep)
    except Exception as ex:
        etype, value, tb = sys.exc_info()
        stack = ''.join(traceback.format_exception(etype,
                                                   value,
                                                   tb,
                                                   100))
        return {
            'message': str(ex),
            'traceback': stack,
            'case': case.__class__.__name__,
        }

    return None


def _alert_exceptions(exceptions):
    if not exceptions:
        return

    alerts = [
        SlackAlert(),
        ConsoleAlert()
    ]

    pool = Pool(len(alerts))
    pool.map(lambda x: x.call(exceptions), alerts)
    pool.join()


def _send_report(title, is_pass, stat):
    reports = [
        SlackAlert(),
        ConsoleAlert()
    ]

    pool = Pool(len(reports))
    pool.map(lambda x: x.report(title, is_pass, stat), reports)
    pool.join()


def cat():
    CONF()
    API = _get_api(CONF)

    start = time.time()
    cases = [
        InstanceCase(),
        KeyPairCase(),
        ImageCase(),
        InstanceTypeCase(),
        JobCase(),
        MonitorCase(),
        NetworkCase(),
        EIPCase(),
        OperationCase(),
        QuotaCase(),
        VolumeCase(),
        SnapshotCase(),
    ]

    exceptions = map(functools.partial(_run_case, api=API), cases)
    exceptions = filter(lambda x: x, exceptions)
    _alert_exceptions(exceptions)

    end = time.time()
    print ('cat finished in %d seconds' % (end - start))


def clean():
    CONF()
    API = _get_api(CONF)

    start = time.time()
    cases = [
        CleanEips(),
        CleanSnapshots(),
        CleanVolumes(),
        CleanInstances(),
        CleanNetworks(),
        CleanImages(),
        CleanKeyPairs(),
    ]

    exceptions = map(functools.partial(_run_case, api=API), cases)
    exceptions = filter(lambda x: x, exceptions)
    _alert_exceptions(exceptions)

    end = time.time()
    print ('clean finished in %d seconds' % (end - start))


def bench():
    CONF()
    API = _get_api(CONF)

    bmconf = getattr(CONF, 'benchmark')
    concurrent = bmconf.concurrent
    total = bmconf.total
    ratio = bmconf.ratio
    timeout = bmconf.timeout

    # begin all, clean them.
    clean()

    cases = [
        BenchVolumes(API, concurrent, gevent.sleep)
    ]

    for case in cases:
        threads = []

        start = time.time()
        t = Timeout(timeout)
        t.start()
        try:
            pool = Pool(concurrent)
            for i in xrange(0, total):
                threads.append(pool.spawn(case.run, API, gevent.sleep))
            pool.join()

        except Timeout:
            pass

        finally:
            t.cancel()

        end = time.time()
        succ = len([t1.value for t1 in threads if t1.value])
        fail = len([t2.value for t2 in threads if not t2.value])
        not_run = total - succ - fail

        title = ('Report: case: %s finished in : %d seconds. \n'
                 'Target: run %d total (with %d concurrent) must '
                 'finish in %d seconds with %d%% success rate.') % (
                  case.__class__.__name__,
                  end - start,
                  total,
                  concurrent,
                  timeout,
                  ratio)

        stat = {
            'total': total,
            'succ': succ,
            'fail': fail,
            'not_run': not_run,
            'time': int(end - start),
        }

        is_pass = int(succ * 100.0 / total) >= ratio

        _send_report(title, is_pass, stat)

    # after all, clean them.
    clean()



import requests
import json


class OneAlert(object):

    def call(self, exceptions):
        for exception in exceptions:
            url = 'http://api.110monitor.com/alert/api/event'
            payload = json.dumps({
                'app': 'bcc67cab-afab-8259-6f86-42402abc52ab',
                'eventId': exception['case'],
                'eventType': 'trigger',
                'alarmName': exception['message'],
                'entityName': exception['case'] + '-name',
                'entityId': exception['case'] + '-id',
                'priority': 1,
                'alarmContent': exception
            })
            requests.post(url, payload)




class BaseException(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class ParamInvalidError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class AccessKeyInvalidError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ResourceUnauthorizedError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ResourceNotFoundError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ActionDisallowError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class RequestFormatInvalidError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class QuotaNotEnoughError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ServerErrorError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ActionsTotallyFailedError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class ActionsPartiallyFailedError(BaseException):
    def __init__(self, message):
        BaseException.__init__(self, message)


class HTTPError(Exception):
    def __init__(self, status_code):
        self.status_code = status_code

    def __str__(self):
        return 'http failed status=%s' % self.status_code



from plato_cat.cases import get_resource


class MonitorCase():

    def run(self, API, sleep):
        instances = get_resource(API, 'DescribeInstances', {
            'status': ['active']
        }, 'instanceSet')

        # delete instances
        instance_ids = [i['instanceId'] for i in instances]

        metrics = [
            'instance.cpu',
            'instance.memory',
            'instance.disk.usage',
            'instance.disk.iops',
            'instance.disk.io',
            'instance.network.traffic',
            'instance.network.packets',
            'volume.usage',
            'volume.iops',
            'volume.io',
            'eip.traffic',
            'eip.packets',
        ]

        for period in [
            '120mins',
            '720mins',
            '48hours',
            '14days',
            '30days',
        ]:
            API.conn.call('GetMonitor', {
                'resourceIds': instance_ids,
                'metrics': metrics,
                'period': period,
            })



from .. import conn


class _API(object):
    """
    actions underlying api
    """
    def __init__(self):
        pass

    def configure(self, access_key, access_secret, endpoint, is_debug):
        """
        """
        if not endpoint.endswith('/'):
            endpoint += '/'

        self.conn = conn.Conn(access_key,
                              access_secret,
                              endpoint,
                              is_debug)

API = _API()


def setup(access_key, access_secret, endpoint, is_debug=False):
    # config this api.
    API.configure(access_key,
                  access_secret,
                  endpoint,
                  is_debug)
    return API



from setuptools import find_packages, setup
from plato_cat import __version__


setup(name="plato_cat",
      version=__version__,
      description="a cat",
      author="piaoyuankui@le.com",
      url="http://plato_cat.com",
      license="MIT",
      install_requires=[
          'requests==2.10.0',
          'gevent==1.1.2',
          'oslo.config==3.7.0',
      ],
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'plato-cat=plato_cat.cmd:cat',
              'plato-clean=plato_cat.cmd:clean',
              'plato-benchmark=plato_cat.cmd:bench'
          ]
      })



from plato_cat.cases import get_resource
from plato_cat.cases import post_resource


class KeyPairCase():

    def run(self, API, sleep):
        keypairs = get_resource(API, 'DescribeKeyPairs', {
            'status': ['active']
        }, 'keyPairSet')

        API.conn.call('CreateKeyPair', {
            'name': 'key-pair-name'
        })

        new_keypairs = get_resource(API, 'DescribeKeyPairs', {
            'status': ['active']
        }, 'keyPairSet')

        if len(new_keypairs) - len(keypairs) != 1:
            raise Exception("after create 2 key_pairs,"
                            " DescribeKeyPairs did not increase by 2!")


class CleanKeyPairs():

    def run(self, API, sleep):
        key_pairs = get_resource(API, 'DescribeKeyPairs', {
            'status': ['active']
        }, 'keyPairSet')

        # delete key_pairs
        key_pair_ids = [kp['keyPairId'] for kp in key_pairs]
        post_resource(API, 'DeleteKeyPairs', key_pair_ids, 'keyPairIds')
